#!/usr/bin/env node

var promisify = require('promise-toolbox').promisify

var camelCase = require('lodash/camelCase')
var join = require('path').join
var map = require('lodash/map')
var readdir = promisify(require('fs').readdir)
var stat = promisify(require('fs').stat)
var writeFile = promisify(require('fs').writeFile)

// ===================================================================

function _log () {
  for (var i = 0, n = arguments.length; i < n; ++i) {
    this.write(arguments[i])
    this.write('\n')
  }
}

function log () {
  _log.apply(process.stdout, arguments)
}

function warn () {
  _log.apply(process.stderr, arguments)
}

function fatal () {
  _log.apply(process.stderr, arguments)
  process.exit(1)
}

// -------------------------------------------------------------------

function removeSuffix (str, sfx) {
  var strLength = str.length
  var sfxLength = sfx.length

  var pos = strLength - sfxLength
  if (pos < 0 || str.indexOf(sfx, pos) !== pos) {
    return false
  }

  return str.slice(0, pos)
}

// ===================================================================

var GENERATORS = (function () {
  var HEADER = [
    '//',
    '// This file has been generated by [index-modules](https://npmjs.com/index-modules)',
    '//',
    ''
  ]

  return {
    default: {
      init: function () {
        return HEADER.concat(
          'const properties = {}',
          ''
        )
      },
      addModule: function (content, path, identifier) {
        content.push(
          'import _' + identifier + ', * as __' + identifier + ' from ' + JSON.stringify(path),
          'properties.' + identifier + ' = { enumerable: true, get: () => _' + identifier + ' }',
          'export { __' + identifier + ' as ' + identifier + ' }',
          ''
        )
      },
      compile: function (content) {
        content.push(
          'export default Object.create(null, properties)',
          ''
        )
        return content.join('\n')
      }
    },
    cjsLazy: {
      init: function () {
        return HEADER.concat(
        'var d = Object.defineProperty',
        'function de (o, n, v) {', // define enumerable property
        '  d(o, n, { enumerable: true, value: v })',
        '  return v',
        '}',
        'function dl (o, n, g, a) {', // define lazy property
        '  d(o, n, { configurable: true, enumerable: true, get: function () { return de(o, n, g(a)) } })',
        '}',
        'function r (p) {', // require module (with ES2015 compat)
        '  var v = require(p)',
        '  return v && v.__esModule ? v',
        "    : typeof v === 'object' || typeof v === 'function' ? Object.create(v, { default: { enumerable: true, value: v } })",
        '    : { default: v }',
        '}',
        'function e (p, i) {', // export a module
        '  dl(defaults, i, function () {',
        '    return exports[i].default',
        '  })',
        '  dl(exports, i, r, p)',
        '}',
        '',
        "d(exports, '__esModule', { value: true })",
        "var defaults = de(exports, 'default', {})"
        )
      },
      addModule: function (content, path, identifier) {
        content.push('e(' + JSON.stringify(path) + ', ' + JSON.stringify(identifier) + ')')
      },
      compile: function (content) {
        content.push(
          ''
        )
        return content.join('\n')
      }
    }
  }
})()

// ===================================================================

var generator

function indexModules (dir) {
  var content = generator.init()

  return readdir(dir).then(function (entries) {
    var index = join(dir, 'index.js')
    return Promise.all(map(entries, function (entry) {
      return entry !== 'index.js' && stat(join(dir, entry)).then(function (stats) {
        var base
        if (stats.isDirectory()) {
          base = entry
        } else if (!(
          stats.isFile() && (
            (base = removeSuffix(entry, '.coffee')) ||
            (base = removeSuffix(entry, '.js'))
          )
        )) {
          return
        }

        generator.addModule(content, './' + base, camelCase(base))
      }, function (error) {
        warn('failed to read ' + dir, error)
      })
    })).then(function () {
      return writeFile(index, generator.compile(content)).then(function () {
        log('index generated ' + index)
      }, function (error) {
        warn('failed to write ' + index, error)
      })
    }, function (error) {
      warn('failed to generate ' + index, error)
    })
  })
}

function findDirs (dir) {
  return readdir(dir).then(function (entries) {
    return Promise.all(map(entries, function (entry) {
      var path = join(dir, entry)
      return stat(path).then(function (stats) {
        if (stats.isDirectory()) {
          return findDirs(path)
        }
        if (entry === '.index-modules' && stats.isFile()) {
          return indexModules(dir)
        }
      }, function (error) {
        warn('cannot read ' + entry, error)
      })
    }))
  }, function (error) {
    warn('failed to read ' + dir, error)
  })
}

// ===================================================================

(function main (args) {
  if (!args.length) {
    return log(
      'Usage: index-modules [--cjs-lazy] <dir>...',
      '       index-modules [--cjs-lazy] --auto <root dir>'
    )
  }

  generator = GENERATORS.default
  if (args[0] === '--cjs-lazy') {
    generator = GENERATORS.cjsLazy
    args.shift()
  }

  if (args[0] === '--auto') {
    if (args.length < 2) {
      fatal('missing param <root dir>')
    }

    findDirs(args[1])
  } else {
    Promise.all(map(args, indexModules))
  }
})(process.argv.slice(2))
